<?php
namespace Undefined\TranslateLocallang\Service;
/***************************************************************
 *  Copyright notice
 *
 *  (c) 2016-2017 Raphael Graf <r@undefined.ch>
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Utility\GeneralUtility;

class XliffService
{
	/**
	* @var array
	*/
	protected $data = [];

	/**
	* @var string
	*/
	protected $extension = NULL;

	/**
	* @var string
	*/
	protected $file = NULL;

	/**
	* @var string
	*/
	protected $useL10n = TRUE;

	/**
	* @var string
	*/
	protected $sourcelang = 'en';

	/**
	* @var array
	*/
	protected $labelcount = [];

	/**
	 * @param string $extension
	 * @param string $file
	 * @param string $sourcelang
	 * @param bool $useL10n
	 * @return void
	 */
	public function init($extension, $file, $sourcelang = 'en', $useL10n = TRUE) {
		$this->extension = $extension;
		$this->file = $file;
		$this->sourcelang = $sourcelang;
		$this->useL10n = $useL10n;
	}

	/**
	 * load data from file(s)
	 *
	 * @param string $langKey
	 * @return bool
	 */
	public function loadLang($langKey) {
		//Note: This does not use TYPO3\CMS\Core\Localization\Parser\XliffParser because of CDATA support
		//load default data first
		if ($this->getLabelCount('default') === NULL) {
			$this->labelcount[$langKey] = 0;
			$this->data = [];
			$fileref = $this->getXlfPath('default', FALSE);
			if(!$this->loadFile($fileref, 'default', TRUE))
				return FALSE;
		}

		//load overlay data
		if ($langKey !== 'default') {
			$this->labelcount[$langKey] = 0;
			$children = [];
			//check for file in typo3conf/ext first
			$fileref = $this->getXlfPath($langKey, FALSE);

			$addkeys = !$this->useL10n;
			$success = $this->loadFile($fileref, $langKey, $addkeys);

			if ($this->useL10n) {
				$fileref = $this->getXlfPath($langKey, TRUE);
				$success = $this->loadFile($fileref, $langKey, TRUE) || $success;
			}

			if (!$success)
				return FALSE;

			//set missing labels to ''
			$count = 0;
			foreach($this->data as $key => $value) {
				if (!isset($value[$langKey]))
					$this->data[$key][$langKey] = '';
				else
					$count++;
			}
			$this->labelcount[$langKey] = $count;
		}

		return TRUE;
	}

	/**
	 * load data from file
	 *
	 * @param string $fileref
	 * @param string $langKey
	 * @param bool $addkeys
	 * @return bool
	 */
	private function loadFile($fileref, $langKey = 'default', $addkeys = TRUE) {
		if (!is_file($fileref))
			return FALSE;
		libxml_use_internal_errors(TRUE);
		$xml = simplexml_load_file($fileref, 'SimpleXMLElement');
		if ($xml === FALSE) {
				echo '<pre>Error parsing file: ' . $fileref . "\n";
				print_r(libxml_get_errors());
				echo '</pre>';
				return FALSE;
		}
		$children = [];
		if ($xml !== FALSE) {
			if (isset($xml->file) && isset($xml->file->body))
				$children = $xml->file->body->children();

			foreach ($children as $transunit) {
				if (isset($transunit['id'])) {
					$key = (string)$transunit['id'];
					$doadd = ($addkeys || isset($this->data[$key]));
					if ($doadd) {
						if (!isset($this->data[$key]))
							$this->data[$key] = [];
						$value = $str = '';
						if ($langKey === 'default' && isset($transunit->source)) {
							$value = (string)$transunit->source;
							$str = $transunit->source->asXML();
						} else if (isset($transunit->target)) {
							$value = (string)$transunit->target;
							$str = $transunit->target->asXML();
						}
						$s = 8;
						while(ctype_space($str[$s]))
							++$s;
						if (isset($str[$s]) && substr_compare($str, '<![CDATA[' , $s, 9) === 0)
							$value = '<![CDATA[' . $value . ']]>';

						if ($value)
							$this->data[$key][$langKey] = $value;
					}
				}
			}
		}
		return TRUE;
	}

	/**
	 * add blank language data
	 *
	 * @param string $langKey
	 * @return void
	 */
	public function addLang($langKey) {
		foreach($this->data as $key => $dummy)
			$this->data[$key][$langKey] = '';
		$this->labelcount[$langKey] = 0;
	}

	/**
	 * @param string $langKey
	 * @return bool
	 */
	public function saveLang($langKey) {
		$labels = []; //matches the format used by TYPO3\CMS\Core\Localization\Parser\XliffParser
		foreach($this->data as $key => $dummy) {
			if (isset($this->data[$key]['default']) && strpos($key, '_newkey') !== 0)
				$labels[$key] = [
					0 => [
						'source' => $this->encodeValue($this->data[$key]['default']),
						'target' => $this->encodeValue($this->data[$key][$langKey])
				]];
		}

		$objectManager = GeneralUtility::makeInstance('TYPO3\\CMS\\Extbase\\Object\\ObjectManager');
		$xliffview = $objectManager->get('TYPO3\\CMS\\Fluid\\View\\StandaloneView');
		$xliffview->setTemplatePathAndFilename(PATH_typo3conf . 'ext/translate_locallang/Resources/Private/Templates/Xliff.html');

		date_default_timezone_set('UTC');
		$xliffview->assignMultiple([
			'labels' => $labels,
			'sourcelang' => $this->sourcelang,
			'targetlang' => ($langKey === 'default') ? NULL: $langKey,
			'productname' => $this->extension,
			'date' => date('Y-m-d\TH:i:s\Z'), //date('c')
		]);

		$xliff = $xliffview->render();
		if ($xliff) {
			$fileref = $this->getXlfPath($langKey, $this->useL10n);
			if ($this->useL10n && !is_dir(dirname($fileref))) {
				$folderCreateMask = (isset($GLOBALS['TYPO3_CONF_VARS']['SYS']['folderCreateMask'])) ?
					octdec($GLOBALS['TYPO3_CONF_VARS']['SYS']['folderCreateMask']) : octdec('2775');
				if (!mkdir(dirname($fileref), $folderCreateMask, TRUE))
					return FALSE;
			}
			if ($handle = @fopen($fileref , 'w')) {
				if (fwrite($handle , $xliff) === FALSE)
					return FALSE;
				if (isset($GLOBALS['TYPO3_CONF_VARS']['SYS']['fileCreateMask']))
					@chmod($fileref, octdec($GLOBALS['TYPO3_CONF_VARS']['SYS']['fileCreateMask']));
				return TRUE;
			}
		}
		return FALSE;
	}

	/**
	 * @param string $str
	 * @return string
	 */
	private function encodeValue($str) {
		if (strpos(ltrim($str), '<![CDATA[') === 0)
			return ltrim($str);
		else
			return htmlspecialchars($str);
	}

	/**
	 * @return array
	 */
	public function &getData() {
		return $this->data; //return a reference to the array
	}

	/**
	 * @param array $data
	 * @return void
	 */
	public function setData($data) {
		$this->data = $data;
		$first = reset($this->data);
		if ($first)
			foreach($first as $langKey => $label)
				$this->labelcount[$langKey] = 0;

		foreach($this->data as $key => $value) {
			foreach($value as $langKey => $label)
				if ($label)
					$this->labelcount[$langKey]++;
		}
	}

	/**
	 * @param string $langKey
	 * @return string
	 */
	public function getFilename($langKey) {
		return $this->getXlfPath($langKey, $this->useL10n);
	}

	/**
	 * @param string $oldKey
	 * @param string $newKey
	 * @return void
	 */
	public function changeKey($oldKey, $newKey) {
		//replace the array key
		if(!isset($this->data[$oldKey]))
			return;

		if ($newKey == '') {
			unset($this->data[$oldKey]);
			return;
		}

		$keys = array_keys($this->data);
		$keys[array_search($oldKey, $keys)] = $newKey;

		$data = array_combine($keys, $this->data);
		$this->data = $data;
	}

	/**
	 * get path to XLF, no checking for file existence, arguments are checked by caller
	 *
	 * @param string $extension
	 * @param string $file
	 * @param string $langKey
	 * @param bool $useL10n
	 * @return string
	 */
	private function getXlfPath($langKey = 'default', $useL10n = FALSE) {
		//get default path
		if ($langKey === 'default')
			return PATH_typo3conf . 'ext/' . $this->extension . '/Resources/Private/Language/' . $this->file;

		//get overlay path
		$file = $langKey . '.' . $this->file;

		if ($useL10n)
			$path = PATH_typo3conf . 'l10n/' . $langKey . '/' . $this->extension . '/Resources/Private/Language/' . $file;
		else
			$path = PATH_typo3conf . 'ext/' . $this->extension . '/Resources/Private/Language/' . $file;
		return $path;
	}

	/**
	 * @param string $langKey
	 * @return int
	 */
	public function getLabelCount($langKey) {
		if (isset($this->labelcount[$langKey]))
			return $this->labelcount[$langKey];
		return NULL;
	}

	/**
	 * @param string $langKey
	 * @return int
	 */
	public function fileExists($langKey) {
		return is_file($this->getXlfPath($langKey, $this->useL10n));
	}
}
